@@ -1,5 +1,8 @@
 /*
  * Copyright (C) Mellanox Technologies Ltd. 2001-2015.  ALL RIGHTS RESERVED.
+ * Copyright (c) 2016      The University of Tennessee and The University
+ *                         of Tennessee Research Foundation.  All rights
+ *                         reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -23,15 +26,15 @@ enum {
 /*
  * UCX tag structure:
  *
- * 01234567 01234567 01234567 01234567 01234567 01234567 01234567 01234567
- *                           |                          |
- *      message tag (24)     |     source rank (24)     |  context id (16)
- *                           |                          |
+ * 01234567 01234567 01234567 01234567 01234567 0123 4567 01234567 01234567
+ *                           |                      |
+ *      message tag (24)     |   source rank (20)   |     context id (20)
+ *                           |                      |
  */
 #define PML_UCX_TAG_BITS                       24
-#define PML_UCX_RANK_BITS                      24
-#define PML_UCX_CONTEXT_BITS                   16
-#define PML_UCX_ANY_SOURCE_MASK                0x800000000000fffful
+#define PML_UCX_RANK_BITS                      20
+#define PML_UCX_CONTEXT_BITS                   20
+#define PML_UCX_ANY_SOURCE_MASK                0x80000000000ffffful
 #define PML_UCX_SPECIFIC_SOURCE_MASK           0x800000fffffffffful
 #define PML_UCX_TAG_MASK                       0x7fffff0000000000ul
 
@@ -86,7 +89,7 @@ enum {
 #define PML_UCX_MESSAGE_RELEASE(_message) \
     { \
         ompi_message_return(*(_message)); \
-        *(_message) = NULL; \
+        *(_message) = MPI_MESSAGE_NULL; \
     }
 
 
@@ -96,7 +99,10 @@ struct pml_ucx_persistent_request {
     unsigned                          flags;
     void                              *buffer;
     size_t                            count;
-    ucp_datatype_t                    datatype;
+    union {
+        ucp_datatype_t                datatype;
+        ompi_datatype_t              *ompi_datatype;
+    } datatype;
     ucp_tag_t                         tag;
     struct {
         mca_pml_base_send_mode_t      mode;
@@ -115,9 +121,21 @@ void mca_pml_ucx_recv_completion(void *request, ucs_status_t status,
 
 void mca_pml_ucx_psend_completion(void *request, ucs_status_t status);
 
+void mca_pml_ucx_bsend_completion(void *request, ucs_status_t status);
+
 void mca_pml_ucx_precv_completion(void *request, ucs_status_t status,
                                   ucp_tag_recv_info_t *info);
 
+void mca_pml_ucx_send_nbx_completion(void *request, ucs_status_t status,
+                                     void *user_data);
+
+void mca_pml_ucx_bsend_nbx_completion(void *request, ucs_status_t status,
+                                      void *user_data);
+
+void mca_pml_ucx_recv_nbx_completion(void *request, ucs_status_t status,
+                                     const ucp_tag_recv_info_t *info,
+                                     void *user_data);
+
 void mca_pml_ucx_persistent_request_complete(mca_pml_ucx_persistent_request_t *preq,
                                              ompi_request_t *tmp_req);
 
@@ -128,18 +146,14 @@ void mca_pml_ucx_request_init(void *request);
 void mca_pml_ucx_request_cleanup(void *request);
 
 
-static inline ucp_ep_h mca_pml_ucx_get_ep(ompi_communicator_t *comm, int dst)
-{
-    return ompi_comm_peer_lookup(comm, dst)->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML];
-}
-
 static inline void mca_pml_ucx_request_reset(ompi_request_t *req)
 {
-    req->req_complete          = false;
+    req->req_complete          = REQUEST_PENDING;
 }
 
-static void mca_pml_ucx_set_send_status(ompi_status_public_t* mpi_status,
-                                        ucs_status_t status)
+__opal_attribute_always_inline__
+static inline void mca_pml_ucx_set_send_status(ompi_status_public_t* mpi_status,
+                                               ucs_status_t status)
 {
     if (OPAL_LIKELY(status == UCS_OK)) {
         mpi_status->MPI_ERROR  = MPI_SUCCESS;
